import enum
from functools import cache

from plantstar_shared import syscon_json
from plantstar_shared.errors import SysconProgrammingError


class SysconType(enum.Enum):
    @classmethod
    def get_type_tuple_by_type_name(the_class, type_name):
        # Make a `type_cache` for all of the types the first time this method is called.  It may seem strange to not use a `cache` decorator here, but because we know of all the
        # types up front, we don't need to search for and cache each item individually, on every call to the method, which would be less efficient than building the cache up in a
        # single loop.
        if not hasattr(the_class, "_type_cache"):
            the_class._type_cache = {the_type.value[0]: the_type.value for the_type in the_class}

        type_tuple = the_class._type_cache.get(type_name, None)

        if type_tuple is None:
            class_name_singular = the_class.__name__.removesuffix("s")
            option_names = [f"\"{option[0]}\"" for option in the_class.get_options()]
            option_name_string = ", ".join(option_names)
            raise SysconProgrammingError(f"Invalid {class_name_singular}: \"{type_name}\".  Options are: [{option_name_string}]")

        return type_tuple

    @classmethod
    @cache
    def get_options(the_class):
        return [item.value for item in the_class]

    @classmethod
    @cache
    def convert_to_json(the_class):
        base_dictionary = {item.name: item.value for item in the_class}
        return syscon_json.dumps(base_dictionary, default=lambda _: "Not Serializeable")



import requests

from django.core.signing import Signer
from plantstar_shared.SysconType import SysconType
from plantstar_shared.errors import InvalidApiRequest, SysconProgrammingError
from plantstar_shared.is_valid_signed_string import is_valid_signed_string
from plantstar_shared.global_definitions import safe_now


def send_get_request(api_type, ip_address, connect_timeout=2, read_timeout=10, data=None, signer_key=None, logger=None, use_https=False):
    return _send_get_post_request_base(
        api_type=api_type, ip_address=ip_address, request_function=requests.get, connect_timeout=connect_timeout, read_timeout=read_timeout, data=data, signer_key=signer_key,
        logger=logger,  use_https=use_https
    )


def send_post_request(api_type, ip_address, connect_timeout=2, read_timeout=10, data=None, signer_key=None, logger=None, use_https=False):
    return _send_get_post_request_base(
        api_type=api_type, ip_address=ip_address, request_function=requests.post, connect_timeout=connect_timeout, read_timeout=read_timeout, data=data, signer_key=signer_key,
        logger=logger, use_https=use_https
    )


def _send_get_post_request_base(*, api_type, ip_address, request_function, connect_timeout=2, read_timeout=10, data=None, signer_key=None, logger=None, use_https=False):
    should_sign = api_type.value[2]

    if should_sign and not signer_key:
        raise SysconProgrammingError("This API call requires a signer_key, but one was not provided")
    elif not should_sign and signer_key:
        raise SysconProgrammingError("This API should not have a signer_key, but one was provided")

    if use_https:
        http_prefix = "https"
    else:
        http_prefix = "http"

    api_type_name = api_type.value[0]

    request_url = f"{http_prefix}://{ip_address}/{api_type_name}/"

    if signer_key:
        signer_timestamp = str(safe_now().timestamp())
        signed_string = Signer(key=signer_key, salt=signer_timestamp).sign(api_type_name)

        if data is None:
            data = {}

        data["signed_string"] = signed_string
        data["signer_timestamp"] = signer_timestamp

    try:
        request = request_function(request_url, timeout=(connect_timeout, read_timeout), data=data)
    except Exception as error:
        error_message = f"The API request \"{api_type_name}\" sent to \"{request_url}\" failed with error: {error}"

        if logger is not None:
            logger.error(error_message)

        raise error

    try:
        data_dictionary = request.json()
    except:
        data_dictionary = {}

    return request.status_code, data_dictionary


def validate_request(*, api_type, signer_key, data):
    should_sign = api_type.value[2]

    signed_string = data.get("signed_string", None)
    signer_timestamp = data.get("signer_timestamp", None)

    if should_sign and (not signed_string or not signer_timestamp):
        raise InvalidApiRequest(f"This API call requires a signed_string and signer_timestamp, but {signed_string} and {signer_timestamp} were provided")
    elif not should_sign and (signed_string or signer_timestamp):
        raise InvalidApiRequest(f"This API call should not have a signed_string or signer_timestamp, but {signed_string} and {signer_timestamp} were provided")

    if signed_string and signer_timestamp:
        api_type_name = api_type.value[0]

        if not is_valid_signed_string(signer_key=signer_key, signed_string=signed_string, unsigned_string=api_type_name, salt=signer_timestamp):
            raise InvalidApiRequest(f"Invalid signing: signer_key: {signer_key}, signed_string: {signed_string}, unsigned_string: {api_type_name}, salt: {signer_timestamp}")


class DataCollectionModuleApiTypes(SysconType):
    GET_DATA_COLLECTION_MODULE_SYSTEM_DISK_USAGE = ("get_data_collection_module_system_disk_usage", "get_data_collection_module_system_disk_usage", True)
    GET_DATA_COLLECTION_MODULE_SYSTEM_INFORMATION = ("get_data_collection_module_system_information", "get_data_collection_module_system_information", True)
    GET_SYSTEM_ERROR_DICTIONARY_LIST = ("get_system_error_dictionary_list", "get_system_error_dictionary_list", True)

    GET_DATA_COLLECTION_MODULE_PROCESS_STATUSES = ("get_data_collection_module_process_statuses", "get_data_collection_module_process_statuses", True)
    COLDBOOT_DATA_COLLECTION_MODULE = ("coldboot_data_collection_module", "coldboot_data_collection_module", True)
    REBOOT_DATA_COLLECTION_MODULE = ("reboot_data_collection_module", "reboot_data_collection_module", True)


class ApuApiTypes(SysconType):
    SET_IS_INITIALIZING_STATUS = ("data_collection_module_manager/set_is_initializing_status", "data_collection_module_manager/set_is_initializing_status", True)
    SET_IS_COLDBOOTING_STATUS = ("data_collection_module_manager/set_is_coldbooting_status", "data_collection_module_manager/set_is_coldbooting_status", True)


class CosmosApiTypes(SysconType):
    SEND_DATA_DICTIONARY = ("send_data_dictionary", "send_data_dictionary", True)



import datetime
import inspect

import pytz
from django.utils.timezone import now

from plantstar_shared import syscon_json


def utc_now():
    return datetime.datetime.now(pytz.utc)


def safe_now():
    """If Django's `now()` throws an exception, we default to `utc_now()`.  Calling Django's `now()` will throw an exception when calling it in non-Django environments."""
    try:
        return now()
    except Exception:
        return utc_now()


def convert_to_json(the_class):
    base_dictionary = dict((key, value) for key, value in inspect.getmembers(the_class()) if not key.startswith("__") and not inspect.isfunction(value))

    return syscon_json.dumps(base_dictionary, default=lambda _: "Not Serializeable")


# TODO: Inherit from SyconType and remove `convert_to_json``
class StatusCodes:
    OK = 200
    MOVED_PERMANENTLY = 301
    FOUND = 302
    BAD_REQUEST = 400
    UNAUTHORIZED = 401
    FORBIDDEN = 403
    INTERNAL_SERVER_ERROR = 500
    NOT_IMPLEMENTED = 501
    SERVICE_UNAVAILABLE = 503
    GATEWAY_TIMEOUT = 504

    @staticmethod
    def convert_to_json():
        return convert_to_json(StatusCodes)
